/**
 * @brief The following test suite tests the various timeouts that can be set and used
 *
 * The goal of this test suite is to make sure the various timeout settings work.
 * Work means they are accurate within a tolerance.
 * This tolerance is defined per test case.
 *
 *
 * The following test cases have not yet been implemented due to the difficult nature
 * in creating test cases around these special edge cases.
 *
 * -w, --write_timeout=TIMEOUT          : Write timeout in miliseconds (DEFAULT 500)
 *  Not sure.
 *
 * -s, --squid_read_timeout=TIMEOUT     : Read timeout in miliseconds (DEFAULT 10000)
 * Not sure. Generally useless feature.
 *
 * -S, --squid_write_timeout=TIMEOUT    : Write timeout in miliseconds (DEFAULT 1000)
 * Not sure.
 *
 * -M, --max_timeouts_num=NUM           : max number of consecutive timeout errors to shutdown connection (DEFAULT 5)
 *  related to the squid_read_timeout and squid_write_timeout
 *  these I do not understand why they are even used...
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <sys/time.h>

#include "CUnit/Basic.h"
#include "jopen.h"
#include "socketserver.h"


struct conninfo {
	int turn;
	int sleep;
	int read_count;
	int disconnect_count;
	struct timeval timestamp;
};
static struct conninfo* conninfo_init()
{
	struct conninfo* ci;
	ci = malloc(sizeof(struct conninfo));
	ci->turn=0;
	ci->sleep=0;
	ci->read_count=0;
	ci->disconnect_count=0;
	return ci;
}

static void conninfo_free(struct conninfo* ci)
{
	free(ci);
}



static void connect_callback(__attribute__ ((unused)) struct client_socket* s)
{

}

static void recv_callback(struct client_socket* s)
{
	if ( strchr(s->rxbuffer, '\n') ) {
		printf("ECHO %s", s->rxbuffer);
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}
}

static void recv_delay_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;
	if ( strchr(s->rxbuffer, '\n') ) {
		printf("ECHO %s", s->rxbuffer);
		ci->read_count++;

		if (ci->turn == 0) {
			if (ci->sleep) {
				sleep(ci->sleep);
			}
			ci->turn = 1;
		} else {
			ci->turn = 0;
		}
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}
}

static void recv_noop_callback(struct client_socket* s)
{
	s->rxlen=0;
}

static void disconnect_callback(__attribute__ ((unused)) struct client_socket* s)
{
}

static void disconnect_count_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;
	ci->disconnect_count++;
	gettimeofday(&ci->timestamp, NULL);
}


static int send_and_get_other(JFILE* fp, char* sbuffer, char* ibuffer, int min, int max)
{
	char sendbuffer[1024];
	char buffer[1024];
	int ret;
	char* ptr;
	struct timeval start;
	struct timeval end;
	struct timeval diff;


	gettimeofday(&start, NULL);
	strcpy(sendbuffer, sbuffer);
	strcat(sendbuffer, " -/- - -\n");

	ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

	int len=0;
	buffer[0]='\0';
	while ( strchr(buffer, '\n') == NULL ) {
		ret=read(fp->from_stdout, buffer+len, 1024);
		if ( ret > 0 ) {
			len+=ret;
			buffer[len]='\0';
		}
	}

	// copy what we are looking for into the sendbuffer
	// reused the sendbuffer
	strcpy(sendbuffer, ibuffer);
	ptr=strchr(sendbuffer, ' ');
	if ( ptr ) {
		*ptr='\0';
	}
	ptr=strchr(buffer, ' ');
	if ( ptr ) {
		*ptr='\0';
	}
	ptr=strchr(buffer, '\n');
	if ( ptr ) {
		*ptr='\0';
	}

	if ( strcmp(buffer, sendbuffer) ) {
		fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
		CU_FAIL("We did not get the response we expected");
		return 0;
	}

	gettimeofday(&end, NULL);

	timersub(&end, &start, &diff);

	int msec = (diff.tv_sec * 1000) + (diff.tv_usec / 1000);

	if (msec < min) {
		CU_FAIL("Time taken was to short");
		return 0;
	}
	if ( msec > max ) {
		CU_FAIL("Time taken was to long");
		return 0;
	}

	return 1;
}

static int send_and_get(JFILE* fp, char* sbuffer, int min, int max)
{
	return send_and_get_other(fp, sbuffer, sbuffer, min, max);
}


/**
 * @brief This tests to make sure we will timeout while trying to connect
 */
void timeout_connect_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11224, connect_callback, recv_callback, disconnect_callback);

	fp = jopen("../redir -t1000 -r500 -w500 -m30000 localhost:11224 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// send and get something
	// this should work
	CU_ASSERT(send_and_get(fp, "http://google.ca", 0, 1000));
	CU_ASSERT(send_and_get(fp, "http://google.ca", 0, 1000));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// this will fail, and we want to measure the time it takes to do so
	// should be the queue flush interval... not based on connect time
	//
	// This should take 3 seconds, so we say between 2 and 4 seconds
	// It may come back right away.
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 2000, 4000));

	// We should be in failure mode now.

	// This should come back right away, we can't connect
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}


/**
 * @brief This will make sure our read will timeout
 */
void timeout_read_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11224, connect_callback, recv_noop_callback, disconnect_callback);

	fp = jopen("../redir -t1000 -r500 -w500 -m30000 localhost:11224 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// This should timeout in about 3 seconds
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 1000, 4000));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}

/**
 * @brief This will make sure our read will timeout, and the next request will timeout right away
 */
void timeout_read_many_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11224, connect_callback, recv_noop_callback, disconnect_callback);

	fp = jopen("../redir -t1000 -r500 -w500 -m30000 localhost:11224 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// This should timeout in about 3 seconds
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 1000, 4000));

	// Since we have no available connections, these should timeout right away
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}


/**
 * @brief This will make sure that when the maximum timeout expires, we will wait again
 */
void timeout_read_max_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11224, connect_callback, recv_noop_callback, disconnect_callback);

	fp = jopen("../redir -t1000 -r500 -w500 localhost:11224 -m 5000 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// This should timeout in about 3 seconds
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 1000, 4000));

	// Since we have no available connections, these should timeout right away
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));

	// sleep for 5 seconds, max timeout was set to 5
	sleep(6);

	// now we should timeout again
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 1000, 4000));

	// Since we have no available connections, these should timeout right away
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}

/**
 * @brief Test the reconnent delay action
 *
 * -d, --reconnect_delay=DELAY          : Delay between attempts to connect when the peer is down (DEFAULT 10000)
 *
 * request (fail)
 * request (fail)
 * startup server
 * request (fail, should not succeed)
 * request (fail, should not succeed)
 * wait for 10 seconds
 * requests should now work
 */
void timeout_reconnect_test()
{
	struct socketserver* svr;
	JFILE* fp;


	fp = jopen("../redir -t1000 -r500 -w500 -m30000 localhost:11224 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		return;
	}

	// This should timeout right away, no server ready
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));

	// These should also timeout right away, no server ready
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));

	// Create a new amazing server
	svr = socketserver_create(11224, connect_callback, recv_callback, disconnect_callback);

	// We still should not be able to connect
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));
	CU_ASSERT(send_and_get_other(fp, "http://google.ca", "http://failcloseurl", 0, 1000));

	// Wait for reconnect timeout to expire
	sleep(11);


	// now we should timeout again
	CU_ASSERT(send_and_get(fp, "http://google.ca", 0, 1000));
	CU_ASSERT(send_and_get(fp, "http://google.ca", 0, 1000));
	CU_ASSERT(send_and_get(fp, "http://google.ca", 0, 1000));
	CU_ASSERT(send_and_get(fp, "http://google.ca", 0, 1000));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}



void timeout_read_less_maxtimeout_test()
{
	struct socketserver* svr;
	struct conninfo* ci;
	JFILE* fp;
	int i;

	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;


	// Create a new amazing server
	svr = socketserver_create(11224, connect_callback, recv_delay_callback, disconnect_count_callback);
	svr->tag = ci = conninfo_init();
	socketserver_tagfree_func(svr, (TAGFREE_CALLBACK)conninfo_free);
	ci->sleep = 2;

	fp = jopen("../redir -t1000 -w500 localhost:11224 -r 1000 -m 3000 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// server delay is bigger than read timeout but less than max timeout
	// Perform 5 tests 
	for(i=0; i < 5; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at newline
		ptr=strchr(sendbuffer, '\n');
		*ptr='\0';

		// truncate buffer at newline;
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	// we expect that redirector didn't disconnect
	if (ci->disconnect_count) {
		printf("Redirector disconnected %d times\n", ci->disconnect_count);
		CU_FAIL("Redirector disconnected");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		jclose(fp);
		return;
	}
	// we expect that all reqeust were send twice
	if (ci->read_count != 2 * 5) {
		printf("We expected %d times send but got %d\n", 2 * 5, ci->read_count);
		CU_FAIL("The send count doesn't match the number we expected");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		jclose(fp);
		return;
	}

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed Delay Less Than Maxtimeout test");

	return;
}

void timeout_read_more_maxtimeout_test()
{
	struct socketserver* svr;
	struct conninfo* ci;
	JFILE* fp;
	int i;

	char sendbuffer[256];
	char buffer[1024];
	int ret;
	char* ptr;

	struct timeval start;
	struct timeval diff;
	int msec;

	// Create a new amazing server
	svr = socketserver_create(11224, connect_callback, recv_delay_callback, disconnect_count_callback);
	svr->tag = ci = conninfo_init();
	socketserver_tagfree_func(svr, (TAGFREE_CALLBACK)conninfo_free);
	ci->sleep = 3;

	fp = jopen("../redir -t1000 -w500 localhost:11224 -r 1000 -m 2000 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// server delay is bigger than max timeout
	// Perform 5 tests 
	for(i=0; i < 5; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);

		gettimeofday(&start, NULL);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at newline
		ptr=strchr(sendbuffer, '\n');
		*ptr='\0';

		// truncate buffer at newline;
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}

		timersub(&ci->timestamp, &start, &diff);
		msec = (diff.tv_sec * 1000) + (diff.tv_usec / 1000);
		printf("Disconnected %d msec\n", msec);

		if ( msec > 4000 || msec < 2000) {
			CU_FAIL("We get wrong disconnect time");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}

	}

	// we expect that redirector disconnected each time
	if (ci->disconnect_count != 5) {
		printf("Redirector disconnected %d times, we expected %d\n", ci->disconnect_count, 5);
		CU_FAIL("Redirector disconnected");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		jclose(fp);
		return;
	}
	// we expect that all reqeust were send twice
	if (ci->read_count != 2 * 5) {
		printf("We expected %d times send but got %d\n", 2 * 5, ci->read_count);
		CU_FAIL("The send count doesn't match the number we expected");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		jclose(fp);
		return;
	}

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed Delay Less Than Maxtimeout test");

	return;
}


/**
 * @brief This setup and creation of this most amazing test
 *
 */
int timeout_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test1
	pSuite = CU_add_suite("Timeout Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Connect timeout test", timeout_connect_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Read timeout test", timeout_read_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Read many request timeout test", timeout_read_many_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Read maximum wait timeout", timeout_read_max_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Test connect reconnect delay", timeout_reconnect_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Test delay less than max timeout", timeout_read_less_maxtimeout_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Test delay greater than max timeout", timeout_read_more_maxtimeout_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	return 0;
}
